// file      : xsd/cxx/tree/generator.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/generator.hxx>

#include <cxx/tree/elements.hxx>

#include <cxx/tree/counter.hxx>
#include <cxx/tree/validator.hxx>
#include <cxx/tree/name-processor.hxx>

#include <cxx/tree/tree-forward.hxx>
#include <cxx/tree/tree-header.hxx>
#include <cxx/tree/tree-inline.hxx>
#include <cxx/tree/tree-source.hxx>

#include <cxx/tree/parser-header.hxx>
#include <cxx/tree/parser-source.hxx>

#include <cxx/tree/stream-header.hxx>
#include <cxx/tree/stream-source.hxx>

#include <cxx/tree/serialization-header.hxx>
#include <cxx/tree/serialization-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>

#include <backend-elements/regex.hxx>
#include <backend-elements/indentation/cxx.hxx>
#include <backend-elements/indentation/sloc.hxx>
#include <backend-elements/indentation/clip.hxx>

#include <cult/containers/set.hxx>

#include <boost/filesystem/fstream.hpp>

#include <cctype>   // std::toupper
#include <iostream>
#include <cstdio>   // std::remove

#include "../../usage.hxx"

using std::endl;
using std::wcerr;

typedef
boost::filesystem::wifstream
WideInputFileStream;

typedef
boost::filesystem::wofstream
WideOutputFileStream;

using namespace XSDFrontend::SemanticGraph;


struct AutoUnlink
{
  AutoUnlink (Path const& file, WideOutputFileStream& os)
      : file_ (file), os_ (os), canceled_ (false)
  {
  }

  ~AutoUnlink ()
  {
    if (!canceled_)
    {
      os_.close ();
      std::remove (file_.native_file_string ().c_str ());
    }
  }

  void
  cancel ()
  {
    canceled_ = true;
  }

private:
  Path file_;
  WideOutputFileStream& os_;
  bool canceled_;
};

namespace CXX
{
  namespace
  {
    Char const copyright[] =
    "// Copyright (C) 2005-2006 Code Synthesis Tools CC\n"
    "//\n"
    "// This program was generated by XML Schema Definition Compiler (XSD)\n"
    "//\n"
    "// This program is free software; you can redistribute it and/or modify\n"
    "// it under the terms of the GNU General Public License version 2 as\n"
    "// published by the Free Software Foundation.\n"
    "//\n"
    "// This program is distributed in the hope that it will be useful,\n"
    "// but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
    "// GNU General Public License for more details.\n"
    "//\n"
    "// You should have received a copy of the GNU General Public License\n"
    "// along with this program; if not, write to the Free Software\n"
    "// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA\n"
    "//\n"
    "// In addition, as a special exception, Code Synthesis Tools CC gives\n"
    "// permission to link this program with the Xerces-C++ library (or with\n"
    "// modified versions of Xerces-C++ that use the same license as Xerces-C++),\n"
    "// and distribute linked combinations including the two. You must obey\n"
    "// the GNU General Public License version 2 in all respects for all of\n"
    "// the code used other than Xerces-C++. If you modify this copy of the\n"
    "// program, you may extend this exception to your version of the program,\n"
    "// but you are not obligated to do so. If you do not wish to do so, delete\n"
    "// this exception statement from your version.\n"
    "//\n"
    "// In addition, Code Synthesis Tools CC makes a special exception for\n"
    "// the Free/Libre and Open Source Software (FLOSS) which is described\n"
    "// in the accompanying FLOSSE file.\n"
    "//\n\n";
  }

  namespace Tree
  {
    namespace CLI
    {
      extern Key char_type                = "char-type";
      extern Key output_dir               = "output-dir";
      extern Key generate_polymorphic     = "generate-polymorphic";
      extern Key generate_serialization   = "generate-serialization";
      extern Key generate_inline          = "generate-inline";
      extern Key generate_ostream         = "generate-ostream";
      extern Key generate_default_ctor    = "generate-default-ctor";
      extern Key generate_from_base_ctor  = "generate-from-base-ctor";
      extern Key namespace_map            = "namespace-map";
      extern Key namespace_regex          = "namespace-regex";
      extern Key namespace_regex_trace    = "namespace-regex-trace";
      extern Key include_with_brackets    = "include-with-brackets";
      extern Key include_prefix           = "include-prefix";
      extern Key root_element_first       = "root-element-first";
      extern Key root_element_last        = "root-element-last";
      extern Key root_element_all         = "root-element-all";
      extern Key root_element_none        = "root-element-none";
      extern Key root_element             = "root-element";
      extern Key hxx_suffix               = "hxx-suffix";
      extern Key ixx_suffix               = "ixx-suffix";
      extern Key cxx_suffix               = "cxx-suffix";
      extern Key hxx_regex                = "hxx-regex";
      extern Key ixx_regex                = "ixx-regex";
      extern Key cxx_regex                = "cxx-regex";
      extern Key hxx_prologue             = "hxx-prologue";
      extern Key ixx_prologue             = "ixx-prologue";
      extern Key cxx_prologue             = "cxx-prologue";
      extern Key prologue                 = "prologue";
      extern Key hxx_epilogue             = "hxx-epilogue";
      extern Key ixx_epilogue             = "ixx-epilogue";
      extern Key cxx_epilogue             = "cxx-epilogue";
      extern Key epilogue                 = "epilogue";
      extern Key hxx_prologue_file        = "hxx-prologue-file";
      extern Key ixx_prologue_file        = "ixx-prologue-file";
      extern Key cxx_prologue_file        = "cxx-prologue-file";
      extern Key prologue_file            = "prologue-file";
      extern Key hxx_epilogue_file        = "hxx-epilogue-file";
      extern Key ixx_epilogue_file        = "ixx-epilogue-file";
      extern Key cxx_epilogue_file        = "cxx-epilogue-file";
      extern Key epilogue_file            = "epilogue-file";
      extern Key parts                    = "parts";
      extern Key parts_suffix             = "parts-suffix";
      extern Key export_symbol            = "export-symbol";
      extern Key show_sloc                = "show-sloc";
      extern Key show_classification      = "show-classification";
    }
  }

  Void Tree::Generator::
  usage ()
  {
    std::wostream& e (wcerr);
    ::CLI::Indent::Clip< ::CLI::OptionsUsage, WideChar> clip (e);

    e << "--char-type <type>" << endl
      << " Use <type> as the base character type. Valid\n"
      << " values are 'char' (default) and 'wchar_t'."
      << endl;

    e << "--output-dir <dir>" << endl
      << " Write generated files to <dir> instead of current\n"
      << " directory."
      << endl;


    e << "--generate-polymorphic" << endl
      << " Generate polymorphism-aware code. Specify this\n"
      << " option if you use substitution groups or xsi:type."
      << endl;

    e << "--generate-serialization" << endl
      << " Generate serialization functions. They convert an\n"
      << " in-memory representation back to XML."
      << endl;

    e << "--generate-inline" << endl
      << " Generate certain functions inline."
      << endl;

    e << "--generate-ostream" << endl
      << " Generate ostream insertion operators."
      << endl;

    e << "--generate-default-ctor" << endl
      << " Generate default constructors even for types that\n"
      << " have required members."
      << endl;

    e << "--generate-from-base-ctor" << endl
      << " Generate from-base constructors."
      << endl;

    e << "--namespace-map <xns>=<cns>" << endl
      << " Map XML Schema namespace <xns> to C++ namespace\n"
      << " <cns>. Repeat this option to specify mapping for\n"
      << " more than one XML Schema namespace."
      << endl;

    e << "--namespace-regex <regex>" << endl
      << " Add the provided regular expression to the list of\n"
      << " regular expressions used to translate XML Schema\n"
      << " namespace names to C++ namespace names. The first\n"
      << " successful substitution is used. The last provided\n"
      << " expression is considered first. Expression is a\n"
      << " perl-like regex in the form /pattern/replacement/."
      << endl;

    e << "--namespace-regex-trace" << endl
      << " Trace the process of applying regular expressions\n"
      << " specified with the --namespace-regex option."
      << endl;

    e << "--include-with-brackets" << endl
      << " Use angle brackets (<>) instead of quotes (\"\") in\n"
      << " generated #include directives for included and\n"
      << " imported schema definitions."
      << endl;

    e << "--include-prefix <prefix>" << endl
      << " Add <prefix> to generated #include directive paths\n"
      << " for included and imported schema definitions."
      << endl;

    e << "--root-element-first" << endl
      << " Treat only the first global element as a document\n"
      << " root."
      << endl;

    e << "--root-element-last" << endl
      << " Treat only the last global element as a document\n"
      << " root."
      << endl;

    e << "--root-element-all" << endl
      << " Treat all global elements as document roots."
      << endl;

    e << "--root-element-none" << endl
      << " Don't treat any global elements as document roots."
      << endl;

    e << "--root-element <element>" << endl
      << " Treat only <element> as a document root. Repeat\n"
      << " this option to specify more than one root element."
      << endl;

    e << "--hxx-suffix <suffix>" << endl
      << " Use <suffix> instead of the default '.hxx' to\n"
      << " construct the name of the header file."
      << endl;

    e << "--ixx-suffix <suffix>" << endl
      << " Use <suffix> instead of the default '.ixx' to\n"
      << " construct the name of the inline file."
      << endl;

    e << "--cxx-suffix <suffix>" << endl
      << " Use <suffix> instead of the default '.cxx' to\n"
      << " construct the name of the source file."
      << endl;

    e << "--hxx-regex <regex>" << endl
      << " Use <regex> to construct the name of the header\n"
      << " file. Expression is a perl regex in the form\n"
      << " /pattern/replacement/."
      << endl;

    e << "--ixx-regex <regex>" << endl
      << " Use <regex> to construct the name of the inline\n"
      << " file. Expression is a perl regex in the form\n"
      << " /pattern/replacement/."
      << endl;

    e << "--cxx-regex <regex>" << endl
      << " Use <regex> to construct the name of the source\n"
      << " file. Expression is a perl regex in the form\n"
      << " /pattern/replacement/."
      << endl;


    // Prologues.
    //
    e << "--hxx-prologue <text>" << endl
      << " Insert <text> at the beginning of the header file."
      << endl;

    e << "--ixx-prologue <text>" << endl
      << " Insert <text> at the beginning of the inline file."
      << endl;

    e << "--cxx-prologue <text>" << endl
      << " Insert <text> at the beginning of the source file."
      << endl;

    e << "--prologue <text>" << endl
      << " Insert <text> at the beginning of each generated\n"
      << " file for which there is no file-specific prologue."
      << endl;


    // Epilogues.
    //
    e << "--hxx-epilogue <text>" << endl
      << " Insert <text> at the end of the header file."
      << endl;

    e << "--ixx-epilogue <text>" << endl
      << " Insert <text> at the end of the inline file."
      << endl;

    e << "--cxx-epilogue <text>" << endl
      << " Insert <text> at the end of the source file."
      << endl;

    e << "--epilogue <text>" << endl
      << " Insert <text> at the end of each generated file\n"
      << " for which there is no file-specific epilogue."
      << endl;


    // Prologue files.
    //
    e << "--hxx-prologue-file <file>" << endl
      << " Insert the content of the <file> at the beginning\n"
      << " of the header file."
      << endl;

    e << "--ixx-prologue-file <file>" << endl
      << " Insert the content of the <file> at the beginning\n"
      << " of the inline file."
      << endl;

    e << "--cxx-prologue-file <file>" << endl
      << " Insert the content of the <file> at the beginning\n"
      << " of the source file."
      << endl;

    e << "--prologue-file <file>" << endl
      << " Insert the content of the <file> at the beginning\n"
      << " of each generated file for which there is no file-\n"
      << " specific prologue file."
      << endl;

    // Epilogue files.
    //
    e << "--hxx-epilogue-file <file>" << endl
      << " Insert the content of the <file> at the end of\n"
      << " the header file."
      << endl;

    e << "--ixx-epilogue-file <file>" << endl
      << " Insert the content of the <file> at the end of\n"
      << " the inline file."
      << endl;

    e << "--cxx-epilogue-file <file>" << endl
      << " Insert the content of the <file> at the end of\n"
      << " the source file."
      << endl;

    e << "--epilogue-file <file>" << endl
      << " Insert the content of the <file> at the end of\n"
      << " each generated file for which there is no file-\n"
      << " specific epilogue file."
      << endl;

    // Misc.
    //
    e << "--parts <num>" << endl
      << " Split generated source code into <num> parts."
      << endl;

    e << "--parts-suffix <suffix>" << endl
      << " Use <suffix> instead of the default '-' to\n"
      << " separate the file name from the part number."
      << endl;

    e << "--export-symbol <symbol>" << endl
      << " Export symbol for Win32 DLL export/import control."
      << endl;

    e << "--show-sloc" << endl
      << " Show the number of generated physical source lines\n"
      << " of code (SLOC)."
      << endl;

    e << "--show-classification" << endl
      << " Show anonymous types that are use to classify more\n"
      << " than one attribute/element."
      << endl;
  }

  Tree::CLI::OptionsSpec Tree::Generator::
  options_spec ()
  {
    CLI::OptionsSpec spec;

    spec.option<CLI::char_type> ().default_value ("char");

    spec.option<CLI::hxx_suffix> ().default_value (".hxx");
    spec.option<CLI::ixx_suffix> ().default_value (".ixx");
    spec.option<CLI::cxx_suffix> ().default_value (".cxx");

    spec.option<CLI::parts> ().default_value (1);
    spec.option<CLI::parts_suffix> ().default_value ("-");

    return spec;
  }


  namespace
  {
    Void
    append (WideOutputFileStream& os,
            NarrowString const& path,
            WideInputFileStream& default_is)
    {
      using std::ios_base;

      if (path)
      {
        WideInputFileStream is (path, ios_base::in | ios_base::binary);

        if (!is.is_open ())
        {
          wcerr << path.c_str () << ": error: unable to open in read mode"
                << endl;

          throw Tree::Generator::Failed ();
        }

        os << is.rdbuf ();

      }
      else if (default_is.is_open ())
      {
        os << default_is.rdbuf ();
        default_is.seekg (0, ios_base::beg);
      }
    }

    Void
    append (WideOutputFileStream& os,
            Cult::Containers::Vector<NarrowString> const& primary,
            Cult::Containers::Vector<NarrowString> const& def)
    {
      Cult::Containers::Vector<NarrowString> const& v (
        primary.empty () ? def : primary);

      for (Containers::Vector<NarrowString>::ConstIterator
             i (v.begin ()), e (v.end ()); i != e; ++i)
      {
        os << i->c_str () << endl;
      }
    }
  }


  Void Tree::Generator::
  generate (Tree::CLI::Options const& ops,
            Schema& schema,
            Path const& file_path)
  {
    using std::ios_base;
    using BackendElements::Regex::perl_s;
    namespace Indentation = BackendElements::Indentation;

    using Cult::Containers::Vector;

    typedef Vector<Path> Paths;
    typedef Vector<Evptr<WideOutputFileStream> > WideOutputFileStreams;
    typedef Vector<Evptr<AutoUnlink> > AutoUnlinks;

    try
    {
      // Do option validation.
      //
      if (ops.value<CLI::parts> () < 1)
      {
        wcerr << "error: invalid value for option --parts: " <<
          ops.value<CLI::parts> () << endl;
        throw Failed ();
      }

      // Get counts.
      //
      Counts counts;
      {
        Counter counter;
        counts = counter.count (ops, schema);

        /*
          wcerr << "global type count: " << counts.global_types << endl;
          wcerr << "global element count: " << counts.global_elements << endl;
          wcerr << "generated global element count: " <<
          counts.generated_global_elements << endl;
        */
      }

      // Evaluate the graph for possibility of generating something useful.
      //
      {
        Validator validator;
        if (!validator.validate (ops, schema, file_path, counts))
          throw Failed ();
      }

      // Process names.
      //
      {
        NameProcessor proc;
        proc.process (schema, file_path);
      }

      // Parts.
      //
      UnsignedLong parts (ops.value<CLI::parts> ());
      UnsignedLong units (
        counts.global_types + counts.generated_global_elements);
      UnsignedLong units_per_part (units / parts);

      if (units != 0 && units_per_part < 1)
      {
        wcerr << "error: too many parts specified: " << parts << endl;
        throw Failed ();
      }

      NarrowString parts_suffix (ops.value<CLI::parts_suffix> ());



      // Generate code.
      //
      NarrowString name (file_path.leaf ());

      NarrowString hxx_suffix (ops.value <CLI::hxx_suffix> ());
      NarrowString ixx_suffix (ops.value <CLI::ixx_suffix> ());
      NarrowString cxx_suffix (ops.value <CLI::cxx_suffix> ());

      NarrowString hxx_expr (ops.value <CLI::hxx_regex> ().empty ()
                             ? "#^(.+?)(\\.[^./\\\\]+)?$#$1" + hxx_suffix + "#"
                             : ops.value <CLI::hxx_regex> ());

      NarrowString ixx_expr (ops.value <CLI::ixx_regex> ().empty ()
                             ? "#^(.+?)(\\.[^./\\\\]+)?$#$1" + ixx_suffix + "#"
                             : ops.value <CLI::ixx_regex> ());

      NarrowString cxx_expr (ops.value <CLI::cxx_regex> ().empty ()
                             ? "#^(.+?)(\\.[^./\\\\]+)?$#$1" + cxx_suffix + "#"
                             : ops.value <CLI::cxx_regex> ());


      NarrowString hxx_name (perl_s (name, hxx_expr));
      NarrowString ixx_name (perl_s (name, ixx_expr));

      Path hxx_path (hxx_name);
      Path ixx_path (ixx_name);
      Paths cxx_paths;

      if (parts > 1)
      {
        for (UnsignedLong i (0); i < parts; ++i)
        {
          std::ostringstream os;
          os << i;

          NarrowString expr (
            "#^(.+?)(\\.[^./\\\\]+)?$#$1" + parts_suffix + os.str () + "$2#");
          NarrowString part_name (perl_s (name, expr));

          part_name = perl_s (part_name, cxx_expr);
          cxx_paths.push_back (part_name);
        }
      }
      else
        cxx_paths.push_back (Path (perl_s (name, cxx_expr)));


      if (NarrowString dir  = ops.value<CLI::output_dir> ())
      {
        try
        {
          Path path (dir, boost::filesystem::native);

          hxx_path = path / hxx_path;
          ixx_path = path / ixx_path;

          for (Paths::Iterator i (cxx_paths.begin ());
               i != cxx_paths.end (); ++i)
            *i = path / *i;
        }
        catch (InvalidPath const&)
        {
          wcerr << dir.c_str () << ": error: invalid path" << endl;
          throw Failed ();
        }
      }

      WideOutputFileStream hxx (hxx_path, ios_base::out);
      WideOutputFileStream ixx;
      WideOutputFileStreams cxx;

      // HXX
      //
      if (!hxx.is_open ())
      {
        wcerr << hxx_path << ": error: unable to open in write mode" << endl;
        throw Failed ();
      }

      AutoUnlink unlink_hxx (hxx_path, hxx);


      // IXX
      //
      Boolean inline_ (ops.value<CLI::generate_inline> ());

      if (inline_)
      {
        ixx.open (ixx_path, ios_base::out);

        if (!ixx.is_open ())
        {
          wcerr << ixx_path << ": error: unable to open in write mode" << endl;
          throw Failed ();
        }
      }

      AutoUnlink unlink_ixx (ixx_path, ixx);

      if (!inline_)
        unlink_ixx.cancel ();


      // CXX
      //
      AutoUnlinks unlink_cxx;

      for (Paths::Iterator i (cxx_paths.begin ());
           i != cxx_paths.end (); ++i)
      {
        Evptr<WideOutputFileStream> s (
          new WideOutputFileStream (*i, ios_base::out));

        if (!s->is_open ())
        {
          wcerr << *i << ": error: unable to open in write mode" << endl;
          throw Failed ();
        }

        unlink_cxx.push_back (Evptr<AutoUnlink> (new AutoUnlink (*i, *s)));
        cxx.push_back (s);
      }


      // Print copyright and license.
      //
      hxx << copyright;
      if (inline_)
        ixx << copyright;

      for (WideOutputFileStreams::Iterator i (cxx.begin ());
           i != cxx.end (); ++i)
        **i << copyright;


      // Prologue.
      //
      WideInputFileStream prologue;
      {
        NarrowString name (ops.value<CLI::prologue_file> ());

        if (!name.empty ())
        {
          prologue.open (name, ios_base::in | ios_base::binary);

          if (!prologue.is_open ())
          {
            wcerr << name.c_str () << ": error: unable to open in read mode"
                  << endl;
            throw Failed ();
          }
        }
      }


      // Epilogue.
      //
      WideInputFileStream epilogue;
      {
        NarrowString name (ops.value<CLI::epilogue_file> ());

        if (!name.empty ())
        {
          epilogue.open (name, ios_base::in | ios_base::binary);

          if (!epilogue.is_open ())
          {
            wcerr << name.c_str () << ": error: unable to open in read mode"
                  << endl;
            throw Failed ();
          }
        }
      }

      // SLOC counter.
      //
      UnsignedLong sloc (0);
      Boolean show_sloc (ops.value<CLI::show_sloc> ());


      // HXX
      //
      {
        Context ctx (hxx, ops, counts);

        Indentation::Clip<Indentation::SLOC, WideChar> hxx_sloc (hxx);

        // Guard
        //
        String guard (hxx_name);
        guard = perl_s (guard, L"/([a-z])([A-Z])/$1_$2/"); // Split words.
        std::transform (guard.begin (), guard.end(), guard.begin (), upcase);
        guard = ctx.escape (guard); // make a c++ id

        hxx << "#ifndef " << guard << endl
            << "#define " << guard << endl
            << endl;


        // Copy prologue.
        //
        hxx << "// Begin prologue." << endl
            << "//" << endl;

        append (
          hxx, ops.value<CLI::hxx_prologue> (), ops.value<CLI::prologue> ());
        append (hxx, ops.value<CLI::hxx_prologue_file> (), prologue);

        hxx << "//" << endl
            << "// End prologue." << endl
            << endl;

        {
          hxx << "#include <xsd/cxx/pre.hxx>" << endl
              << endl;

          // Set auto-indentation.
          //
          Indentation::Clip<Indentation::CXX, WideChar> hxx_clip (hxx);


          // Generate.
          //
          generate_forward (ctx, schema);

          //@@ move expr to ctx
          //
          generate_tree_header (ctx, schema, hxx_expr, ixx_expr);

          generate_parser_header (ctx, schema);

          if (ops.value<CLI::generate_ostream> ())
            generate_stream_header (ctx, schema);

          if (ops.value<CLI::generate_serialization> ())
            generate_serialization_header (ctx, schema);

        /*
          if (traversal)
          generate_traversal_header (ctx, schema);

          if (writer)
          generate_writer_header (ctx, schema);
        */

          if (inline_)
          {
            hxx << "#ifndef XSD_DONT_INCLUDE_INLINE" << endl
                << endl
                << "#include \"" << ixx_name.c_str () << "\"" << endl
                << endl
                << "#endif // XSD_DONT_INCLUDE_INLINE" << endl
                << endl;
          }

          hxx << "#include <xsd/cxx/post.hxx>" << endl
              << endl;
        }

        // Copy epilogue.
        //
        hxx << "// Begin epilogue." << endl
            << "//" << endl;

        append (hxx, ops.value<CLI::hxx_epilogue_file> (), epilogue);
        append (
          hxx, ops.value<CLI::hxx_epilogue> (), ops.value<CLI::epilogue> ());

        hxx << "//" << endl
            << "// End epilogue." << endl
            << endl;

        hxx << "#endif // " << guard << endl;

        if (show_sloc)
        {
          wcerr << hxx_path << ": "
                << hxx_sloc.buffer ().count () << std::endl;

          sloc += hxx_sloc.buffer ().count ();
        }
      }


      // IXX
      //
      if (inline_)
      {
        Context ctx (ixx, ops, counts);

        Indentation::Clip<Indentation::SLOC, WideChar> ixx_sloc (ixx);

        // Guard
        //
        String guard (ixx_name);
        guard = perl_s (guard, L"/([a-z])([A-Z])/$1_$2/"); // Split words.
        std::transform (guard.begin (), guard.end(), guard.begin (), upcase);
        guard = ctx.escape (guard); // make a c++ id

        ixx << "#ifndef " << guard.c_str () << endl
            << "#define " << guard.c_str () << endl
            << endl;

        // Copy prologue.
        //
        ixx << "// Begin prologue." << endl
            << "//" << endl;

        append (
          ixx, ops.value<CLI::ixx_prologue> (), ops.value<CLI::prologue> ());
        append (ixx, ops.value<CLI::ixx_prologue_file> (), prologue);

        ixx << "//" << endl
            << "// End prologue." << endl
            << endl;

        {
          // Set auto-indentation.
          //
          Indentation::Clip<Indentation::CXX, WideChar> ixx_clip (ixx);


          // Generate.
          //
          generate_tree_inline (ctx, schema, ixx_expr, 1, 0);
        }

        // Copy epilogue.
        //
        ixx << "// Begin epilogue." << endl
            << "//" << endl;

        append (ixx, ops.value<CLI::ixx_epilogue_file> (), epilogue);
        append (
          ixx, ops.value<CLI::ixx_epilogue> (), ops.value<CLI::epilogue> ());

        ixx << "//" << endl
            << "// End epilogue." << endl
            << endl;

        ixx << "#endif // " << guard.c_str () << endl;

        if (show_sloc)
        {
          wcerr << ixx_path << ": "
                << ixx_sloc.buffer ().count () << std::endl;

          sloc += ixx_sloc.buffer ().count ();
        }
      }


      // CXX
      //


      for (UnsignedLong unit (0); unit < parts; ++unit)
      {
        UnsignedLong first (unit * units_per_part);
        UnsignedLong last ((unit + 1 == parts)
                           ? (units - 1)
                           : (first + units_per_part - 1));

        // wcerr << "[" << first << ", " << last << "]" << endl;

        WideOutputFileStream& os (*cxx[unit]);

        Context ctx (os, ops, counts);

        Indentation::Clip<Indentation::SLOC, WideChar> cxx_sloc (os);

        // Copy prologue.
        //
        os << "// Begin prologue." << endl
           << "//" << endl;

        append (os,
                ops.value<CLI::cxx_prologue> (), ops.value<CLI::prologue> ());
        append (os, ops.value<CLI::cxx_prologue_file> (), prologue);

        os << "//" << endl
           << "// End prologue." << endl
           << endl;

        {
          os << "#include <xsd/cxx/pre.hxx>" << endl
             << endl;

          // Set auto-indentation.
          //
          Indentation::Clip<Indentation::CXX, WideChar> cxx_clip (os);


          // Generate.
          //
          os << "#include \"" << hxx_name.c_str () << "\"" << endl
             << endl;

          if (!inline_)
            generate_tree_inline (ctx, schema, L"", first, last);

          generate_tree_source (ctx, schema, first, last);

          generate_parser_source (ctx, schema, first, last);

          if (ops.value<CLI::generate_ostream> ())
            generate_stream_source (ctx, schema, first, last);

          if (ops.value<CLI::generate_serialization> ())
            generate_serialization_source (ctx, schema, first, last);

          os << "#include <xsd/cxx/post.hxx>" << endl
             << endl;
        }

        // Copy epilogue.
        //
        os << "// Begin epilogue." << endl
           << "//" << endl;

        append (os, ops.value<CLI::cxx_epilogue_file> (), epilogue);
        append (os,
                ops.value<CLI::cxx_epilogue> (), ops.value<CLI::epilogue> ());

        os << "//" << endl
           << "// End epilogue." << endl
           << endl;

        if (show_sloc)
        {
          wcerr << cxx_paths[unit] << ": "
                << cxx_sloc.buffer ().count () << std::endl;

          sloc += cxx_sloc.buffer ().count ();
        }

      }

      unlink_hxx.cancel ();
      unlink_ixx.cancel ();

      for (AutoUnlinks::Iterator i (unlink_cxx.begin ());
           i != unlink_cxx.end (); ++i)
        (**i).cancel ();

      if (show_sloc)
        wcerr << "total: " << sloc << std::endl;

    }
    catch (NoNamespaceMapping const& e)
    {
      wcerr << e.file () << ":" << e.line () << ":" << e.column ()
            << ": error: unable to map XML Schema namespace '" << e.ns ()
            << "' to C++ namespace" << endl;

      wcerr << e.file () << ":" << e.line () << ":" << e.column ()
            << ": info: use the --namespace-map or --namespace-regex option "
            << "to provide custom mapping" << endl;

      throw Failed ();
    }
    catch (InvalidNamespaceMapping const& e)
    {
      wcerr << "error: invalid XML to C++ namespace mapping specified: "
            << "'" << e.mapping () << "': " << e.reason () << endl;

      throw Failed ();
    }
  }
}
